from typing import List
import sys
from kslurm.args import parse_args
from kslurm.models import KslurmModel
from kslurm.submission import kjupyter, krun, kbatch
from kslurm.installer import update

def kslurm(script: str, args: List[str]) -> None:
    parsed = parse_args(sys.argv[1:], KslurmModel())
    command = parsed.command.value
    tail = parsed.tail
    if command == "krun":
        krun(command, tail.values)
    if command == "kbatch":
        kbatch(command, tail.values)
    if command == "kjupyter":
        kjupyter(command, tail.values)
    if command == "update":
        update(tail.values)

def main():
    kslurm(sys.argv[0], sys.argv[1:])

if __name__ == "__main__":
    main()


from kslurm.exceptions import CommandLineError
from typing import Any, Callable, List, TypeVar, cast, Iterable, Tuple

import functools as fc
import itertools as it

import kslurm.args.helpers as helpers
from .arg_types import Arg, FlagArg, ShapeArg, KeywordArg, PositionalArg, TailArg

T = TypeVar("T")
S = TypeVar("S")

def parse_args(args: Iterable[str], models: T) -> T:
    try:
        return _parse_args(args, models)
    except CommandLineError as err:
        print(err.msg)
        exit()

def _parse_args(args: Iterable[str], models: T) -> T:
    model_list = helpers.get_arg_list(models)

    model_cats = helpers.group_by_type(model_list)

    specific_args = list(it.chain(
        model_cats.get(ShapeArg, []),
        model_cats.get(KeywordArg, []),
        model_cats.get(FlagArg, [])
    ))
    
    labelled = [classify_arg(a, specific_args) for a in args]
    grouped = helpers.group_by_type(group_keywords(labelled))
    shape_args = grouped.get(ShapeArg, [])
    keyword_args = grouped.get(KeywordArg, [])
    flag_args = grouped.get(FlagArg, [])

    nonspecific_args = cast(
        Iterable[PositionalArg[Any]], 
        grouped.get(PositionalArg, [])
    )

    positional_models = cast(
        Iterable[PositionalArg[Any]],
        filter(lambda x: isinstance(x, PositionalArg), model_list)
    )
    positional_args, extras = delineate_positional(nonspecific_args, positional_models)

    if TailArg in model_cats:
        tail_arg_list = cast(List[TailArg], list(model_cats[TailArg]))
        if len(tail_arg_list) > 1:
            raise CommandLineError("More than 1 TailArgs provided:"
                            f"{tail_arg_list}")
        
        tail_arg = tail_arg_list[0]
        tail = [tail_arg.add_values([v.raw_value for v in extras])]
    else:
        tail = cast(List[TailArg], [])
        if extras:
            raise CommandLineError(f"{extras} does not match any Shape, Keyword, or Positional Args ")
    
    return helpers.update_model(it.chain(shape_args, positional_args, keyword_args, flag_args, tail), models)

def classify_arg(arg: str, arg_list: Iterable[Arg[Any]]):
    for argtype in arg_list:
        if argtype.match(arg):
            return argtype.set_value(arg)
    return cast(Arg[Any], PositionalArg().set_value(arg))


def group_keywords(args: List[Arg[Any]]) -> Iterable[Arg[Any]]:
    return fc.reduce(
        group_keyword, 
        args, 
        ( 0, cast(List[Arg[Any]], []) )
    )[1]
    

def get_keyword_group_size(arg: Arg[Any]):
    if isinstance(arg, KeywordArg):
        return arg.num
    else:
        return 0

def group_keyword(accumulant: Tuple[int, Iterable[Arg[Any]]], arg: Arg[Any]) -> "Tuple[int, Iterable[Arg[Any]]]":
    group_size, running_list = accumulant
    new_keyword_group_size = get_keyword_group_size(arg)

    kcast: Callable[[Any], KeywordArg[Any]] = lambda x: cast(KeywordArg[Any], x)
    l = list(running_list)
    if l:
        last_arg = l[-1]
    else:
        last_arg = None
   
    if group_size:
        assert isinstance(last_arg, KeywordArg)
        kcast(last_arg)
        return (
            group_size - 1, 
            it.chain(
                l[:-1], 
                [kcast(last_arg.add_values(it.chain(
                    last_arg.values,
                    [arg.raw_value]
                )))]
            )
        )
    elif isinstance(last_arg, KeywordArg) and \
        get_keyword_group_size(kcast(last_arg)) == 0 and \
        not isinstance(arg, ShapeArg) and \
        not isinstance(arg, KeywordArg):
        return (
            group_size, 
            it.chain(
                l[:-1], 
                [kcast(last_arg.add_values(it.chain(
                    last_arg.values,
                    [arg.raw_value]
                )))] 
            )
        )
    else:
        return new_keyword_group_size, it.chain(l, [cast(Arg[Any], arg)]) 


def delineate_positional(args: Iterable[PositionalArg[Any]], positional_models: Iterable[PositionalArg[Any]]) -> Tuple[List[PositionalArg[Any]], List[PositionalArg[Any]]]:
    args = list(args)
    models = list(positional_models)
    if len(args) < len(models):
        raise CommandLineError("Too few positional arguments received.\n"
                        f"Expected {models}\n"
                        f"but got {args}")
    
    positional_args = [
        model.set_value(arg.raw_value)
        for arg, model in zip(args, models)
    ]
    
    # We return all the args not zipped with the models as
    # extras to be added to TailArg
    return positional_args, args[len(models):]


from typing import List
import sys, subprocess, re, signal
from kslurm.slurm import SlurmCommand
from kslurm.models import SlurmModel
from kslurm.args import print_help
from kslurm import text as txt
from kslurm.style.console import console



def kjupyter(script: str, args: List[str]):
    """Start a Jupyter session
    
    Begins a Jupyter session on an interactive node. By default, it requests 3hr. No
    more time than this should be requested, as this could significantly delay the 
    start of the server.  
    
    For the command to work, a virtualenv containing jupyter-lab should already be
    activated. Use `pip install jupyter-lab`
    """
    slurm = SlurmCommand(args, SlurmModel())
    if slurm.help:
        print_help(script, SlurmModel(), kjupyter.__doc__) # type: ignore
        exit()
    
    slurm.command = ['jupyter-lab', "--ip", "$(hostname -f)", "--no-browser"]
    print(txt.KRUN_CMD_MESSAGE.format(args=slurm.slurm_args, command=slurm.command))

    if not slurm.test:
        
        proc = subprocess.Popen(slurm.run, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        jobid=0
        
        def signal_handler(*args): # type: ignore
            subprocess.run(f"scancel {jobid}", shell=True)
            sys.exit(0)
        signal.signal(signal.SIGINT, signal_handler) # type: ignore

        while proc.poll() is None:
            if proc.stdout:
                line = proc.stdout.readline().decode().strip()
                queued = re.match(r"^srun: job (\d+) queued and waiting for resources", line)
                url = re.match(r"^https?:\/\/((?:[\w-]+\.)+(?:[\w]+)):(\d+)(\/lab\?token=\w+)$", line)
                if queued:
                    jobid=queued.group(1)
                elif any([
                    re.match(r"^srun: job \d+ has been allocated resources", line),
                    re.match(r"^To access the server, open this file in a browser:$", line),
                    re.match(r"^file:\/\/\/(?:[\w.\-_]+\/)+[\w.\-_]+\.html$", line),
                    re.match(r"^Or copy and paste one of these URLs:$", line),
                    re.match(r"^or http:\/\/(?:[\d]+\.)+\d+:\d+\/lab\?token=\w+$", line)
                ]):
                    pass
                elif url:
                    domain = url.group(1)
                    port = url.group(2)
                    path = url.group(3)
                    console.print(txt.JUPYTER_WELCOME.format(
                        port=port,
                        domain=domain,
                        path=path,
                        url=url.group(0)
                    ))
                else:
                    print(line)

        
        

def main():
    kjupyter(sys.argv[0], sys.argv[1:])

if __name__ == "__main__":
    main()


import subprocess
from typing import List
from venv import EnvBuilder
from pathlib import Path

def install(
        version: str, 
        specification: str, 
        data_dir: Path, 
        bin_dir: Path, 
        update: bool=False, 
        entrypoints: List[str] = []
    ):
    """
    Installs Software.
    """
    try:
        data_dir.mkdir(parents=True, exist_ok=True)
        bin_dir.mkdir(parents=True, exist_ok=True)
        env_path = make_env(data_dir, update)
        install_library(specification, env_path)
        if entrypoints:
            make_bin(bin_dir, data_dir, entrypoints)

        data_dir.joinpath("VERSION").write_text(version)

        return 0
    except subprocess.CalledProcessError as e:
        print(
            f"\nAn error has occurred: {e}\n{e.stderr}"
        )
        return e.returncode

def make_env(data_dir: Path, update: bool = False) -> Path:
    env_path = data_dir / "venv"
    if update and is_venv(env_path):
        return env_path
    print("Making virtual env")
    EnvBuilder(with_pip=True, clear=True).create(str(env_path))
    return env_path

def is_venv(venv: Path) -> bool:
    python = venv / "bin/python"
    pip = venv / "bin/pip"
    # If pip and python exist, that's enough to install something,
    # so we assume it's a venv
    if python.exists() and pip.exists():
        return True
    return False

def install_library(specification: str, env_path: Path) -> None:
    print("Installing")
    python = env_path.joinpath("bin/python")

    subprocess.run(
        [str(python), "-m", "pip", "install", specification],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=True,
    )

def make_bin(bin_dir: Path, data_dir: Path, entrypoints: List[str]) -> None:
    bin_dir.mkdir(parents=True, exist_ok=True)

    for script in entrypoints:
        target_script = "venv/bin/" + script
        if bin_dir.joinpath(script).is_symlink():
            bin_dir.joinpath(script).unlink()

        bin_dir.joinpath(script).symlink_to(
            data_dir.joinpath(target_script)
        )


from typing import List
import sys, subprocess
from colorama import Fore
from kslurm.slurm import SlurmCommand
from kslurm.models import SlurmModel
from kslurm.args import print_help
from .. import text as txt

def kbatch(script: str, args: List[str]):
    """Submit a job using sbatch

    Supports scripts (e.g. ./script.sh) or direct commands (e.g. cp dir/file.txt dir2)

    When your command contains bash interpreted elements such as $VARIABLES and 
    $(subshells), these will be immediately expanded. Normally, this behaviour
    is fine, but sometimes they should only be interpretted on the allocated cluster.
    For instance, $SLURM_TMPDIR will evaluate to "" unless it is interpretted later.
    To force this behaviour, wrap the $VARIABLE or $(subshell) in quotes:
        '$SLURM_TMPDIR'
        '$(hostname)'
    """

    slurm = SlurmCommand(args, SlurmModel())
    if slurm.help:
        print_help(script, SlurmModel(), kbatch.__doc__) # type: ignore
        exit()
    command = slurm.command if slurm.command else f"{Fore.RED}Must provide a command"

    print(txt.KBATCH_MSG.format(slurm_args=slurm.slurm_args, command=command))
    if slurm.command:
        proc = subprocess\
            .run(slurm.batch, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        out = proc.stdout.decode()
    
        if proc.returncode != 0:
            print(Fore.WHITE + out)
            return proc.returncode
        if slurm.test:
            # output will be the issued command, so we print it
            print(Fore.WHITE + out)
        else:
            # We subtract the last character of the output
            # to remove the final "\n" character and get the 
            # job_id
            slurmid = out[:-1]

            print(f"""Scheduled job {slurmid}.
    To cancel, run:
        scancel {slurmid}
        """)

def main():
    kbatch(sys.argv[0], sys.argv[1:])


from __future__ import annotations
from typing import Callable, Generic, Iterable, List, Optional, TypeVar

import  abc, copy

from colorama import Fore, Style

from kslurm.exceptions import CommandLineError, ValidationError

T = TypeVar("T")

class Arg(abc.ABC, Generic[T]):
    raw_value: str

    def __init__(self, *,
            id: str,
            match: Callable[[str], bool],
            value: Optional[str],
            format: Callable[[str], T],
            help: str):
        self.id = id
        self.match = match
        self._format = format
        self.value = value
        self.help = help

    @property
    def value(self):
        if self._value is None:
            raise Exception()
        return self._value
        
    @value.setter
    def value(self, value: Optional[str]):
        if value is None:
            self._value = value
            self.raw_value = ""
        else:
            self._value = self._format(value)
            self.raw_value = value

    def __eq__(self, o: object) -> bool:
        if isinstance(o, self.__class__):
            return o.value == self.value
        else:
            return False
    
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.id} = {self._value}>"

    def __str__(self) -> str:
        return str(self.value)

    def setid(self, id: str):
        c = copy.copy(self)
        c.id = id
        return c

    @abc.abstractmethod
    def set_value(self, value: str) -> Arg[T]:
        pass

class PositionalArg(Arg[T]):
    def __init__(self,
            value: Optional[str] = None, 
            id: str="positional",
            format: Callable[[str], T]=str,
            validator: Callable[[str], str] = lambda x: x,
            help: str = "",
            name: str = ""):
        super().__init__(id=id, match=lambda x: True, value=value, format=format, help=help)
        self.validator = validator
        self.name = name
        self.updated = False

    def set_value(self, value: str):
        try:
            c = copy.copy(self)
            c.value = self.validator(value)
            c.updated = True
            return c
        except ValidationError as err:
            raise CommandLineError(f"""
    {Fore.RED + Style.BRIGHT}ERROR:{Style.RESET_ALL}
        Invalid value for "{Style.BRIGHT + self.name + Style.RESET_ALL}":
            {err.msg}
            """, err)

class ChoiceArg(PositionalArg[T]):
    def __init__(self, *,
            value: Optional[str] = None, 
            match: List[str] ,
            id: str="positional",
            format: Callable[[str], T]=str,
            help: str = "",
            name: str = ""):

        def check_match(val: str):
            if val in match:
                return val
            choices = "\n".join([f"\t\t• {m}" for m in match])
            raise ValidationError(
                f"Please select between:\n"
                f"{choices}"
            )
        
        super().__init__(value=value, id=id, format=format, validator=check_match, help=help, name=name)

        self.match_list = match


class ShapeArg(Arg[T]):
    def __init__(self,*,
            id: str = "", 
            match: Callable[[str], bool],
            value: Optional[str] = None,
            format: Callable[[str], T]=str,
            help: str = "",
            name: str = "",
            syntax: str = "",
            examples: List[str] = []):
        super().__init__(id=id, match=match, value=value, format=format, help=help)
        self.updated = False
        self.name = name
        self.syntax = syntax
        self.examples = examples

    def set_value(self, value: str):
        c = copy.copy(self)
        c.value = value
        c.raw_value = value
        c.updated = True
        return c

class FlagArg(Arg[bool]):
    def __init__(self, *,
            id: str = "",
            match: List[str],
            value: bool = False,
            help: str = ""):

        def check_match(val: str):
            if val in match:
                return True
            return False
        if value:
            raw_value = match[0]
        else:
            raw_value = ""

        super().__init__(id=id, match=check_match, format=bool, value=raw_value, help=help)
        self.match_list = match


    def set_value(self, value: str):
        c = copy.copy(self)
        c.value = value
        return c

S = TypeVar("S")

class KeywordArg(FlagArg, Generic[S]):
    def __init__(self, *,
            id: str = "",
            match: List[str],
            value: bool = False,
            num: int = 1,
            validate: Callable[[str], str] = lambda x: x,
            values: List[str] = [],
            help: str = "",
            values_name: str = ""):
        super().__init__(id=id, match=match, help=help, value=value)
        self.num = num
        self.validate = validate
        self.values = values
        self.values_name = values_name

    @property
    def values(self) -> List[str]:
        return self._values
    
    @values.setter
    def values(self, values: List[str]):
        self._values = values
    
    def set_value(self, value: str):
        c = copy.copy(self)
        c.value = value
        # We need to obliterate any values in order for argparse to work
        # Otherwise any default args will be included in the final list.
        # A more elegant solution would delete default args organically
        # if user-supplied args are added.
        c.values = []
        return c

    def add_values(self, values: Iterable[str]):
        try:
            c = copy.copy(self)
            c.values = list(map(self.validate, values))
            return c
        except ValidationError as err:
            raise CommandLineError(f"""
    {Fore.RED + Style.BRIGHT}ERROR:{Style.RESET_ALL}
        Invalid value for "{Style.BRIGHT + self.id + Style.RESET_ALL}":
            {err.msg}
            """, err) from err


    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.id} = {self.values}>"
    
    def __str__(self) -> str:
        return " ".join(self.values)

    def __eq__(self, o: object) -> bool:
        if isinstance(o, self.__class__):
            return o.value == self.value and o.values == self.values
        else:
            return False


class TailArg(KeywordArg[str]):
    def __init__(self, name: str = "Tail"):
        super().__init__(id="tail", num=-1, match=[])
        self.name = name




import sys, subprocess
from pathlib import Path
import itertools as it

import attr
from colorama import Fore, Style

from kslurm.args.arg_types import PositionalArg
from kslurm.exceptions import ValidationError
from kslurm.slurm import SlurmCommand
from kslurm.models import SlurmModel

# Helper function for formatting the list of settings in our output
def setting_list(name: str, setting: str) -> str:
    return Fore.YELLOW + name + ": " + Fore.WHITE + Style.BRIGHT + setting + Style.RESET_ALL

def profile_validator(profile: str) -> str:
    profile_path = Path.home() / ".config" / "snakemake"
    configfiles = [profile_path.glob(f'*/config.{ext}') for ext in ['yml', 'yaml']]
    profiles = [configfile.parent.name for configfile in it.chain(*configfiles)]
    if profile in profiles:
        return profile
    if profiles:
        profiles = "\n".join(profiles)
        profilelist = f"Found the following profiles:\n{profiles}"
    else:
        profilelist = f"Did not find any valid profiles in {profile_path}"
    raise ValidationError(f"{Fore.RED}\"{Fore.LIGHTRED_EX + profile + Fore.RED}\" "
                            f"is not a valid profile.{Fore.RESET} \n\n{profilelist}")

# Extended Model
@attr.s(auto_attribs=True)
class SSnakeModel(SlurmModel):
    profile: PositionalArg[str] = PositionalArg(
        validator=profile_validator
    )

def main():
    models = SSnakeModel()
    models.cpu.value = "2"
    models.profile 
    slurm = SlurmCommand(sys.argv[1:], models)

    # Get the profile
    profile = slurm.args.profile.value
    

    # Use parent directory name as the job name
    slurm.name = Path.cwd().name
    slurm.output = "snakemake-%j.out"

    # Update our submission script
    slurm.script = [
        "source $SNAKEMAKE_VENV_DIR/activate",
        "panoptes --ip $(hostname -f) 1> panoptes.out 2>&1 &",
        "PANOPTES_PID=$!",
        "(tail -F panoptes.out & ) | grep -q \"Running on\"",
        "hostname -f",
        f"snakemake --wms-monitor \"http://$(hostname -f):5000\" --profile {profile} {slurm.command}",
        "kill $PANOPTES_PID",
        "rm panoptes.out"
    ]

    # Run the process and collect the jobid output. 
    output = subprocess\
        .run(slurm.batch, shell=True, capture_output=True)\
        .stdout.decode()

    if slurm.test:
        # output will be the issued command, so we print it
        print(Fore.WHITE + output)
    else:
        # We subtract the last 2 characters of the output
        # to remove the final "\n" characters and get the 
        # job_id
        slurmid = output[:-2]

        # Print a helpful confirmation message
        print(f"""
    {Fore.GREEN}Scheduling Snakemake
        {Fore.LIGHTBLUE_EX}SETTINGS
            {Fore.WHITE}{slurm.slurm_args}

            {setting_list("profile", profile)}
            {setting_list("job_name", slurm.name)}
            {setting_list("job_id", slurmid)}
            {setting_list("other_args", slurm.command)}
    
    To cancel the job, run:
        scancel {slurmid}
        """) 

if __name__ == "__main__":
    main()


from typing import Iterable, Dict, List, DefaultDict, TypeVar, cast, Any, ItemsView
import itertools as it
from collections import defaultdict
import attr

from .arg_types import Arg

T = TypeVar("T")
C = TypeVar("C", bound=Arg[Any])


def group_by_type(items: Iterable[T]) -> Dict[type, List[T]]:
    groupedDict = cast(DefaultDict[type, List[T]], defaultdict(list))
    for key, value in it.groupby(items, type):
        groupedDict[key] += list(value)
    return groupedDict


def get_arg_list(models: object):
    if hasattr(models, "__attrs_attrs__"):
        return [
            cast(Arg[Any], arg.setid(name)) for name, arg in attr.asdict(models, recurse=False).items()
        ]
    elif isinstance(models, dict):
        return [
            arg.setid(name) for name, arg in cast(ItemsView[str, Arg[Any]], models.items() )
        ]
    else:
        raise Exception(f"{type(models)} is not a supported object for arg models")


def update_model(arg_updates: Iterable[Arg[Any]], model: T) -> T:
    update = { 
        arg.id: arg for arg in arg_updates
    }
    return model.__class__(**update)


import re
from pathlib import Path

import attr

from kslurm.args import ShapeArg, KeywordArg, FlagArg, TailArg
import kslurm.models.formatters as formatters
import kslurm.models.validators as validators

@attr.s(auto_attribs=True)
class SlurmModel:
    time: ShapeArg[int] = ShapeArg[int](
        match = lambda arg: bool(re.match(r'^([0-9]{1,2}-)?[0-9]{1,2}:[0-9]{2}$', arg)),
        format = formatters.time,
        value = "03:00",
        name="Time",
        syntax="[d-]dd:dd",
        help="Amount of time requested. Written as [days-]hr:min.",
        examples=[
            "3:00 (3hr)",
            "1-00:00 (1 day)"
        ])

    gpu: FlagArg = FlagArg(
        match = ["gpu"],
        help="Request a gpu instance")

    cpu: ShapeArg[int] = ShapeArg[int](
        match = lambda arg: bool(re.match(r'^[0-9]+$', arg)),
        value = "1",
        format = int,
        name = "Number of CPUs",
        syntax="d",
        help="Number of CPUs to request. Maximum of 32 on Graham Nodes.",
        examples=[
            "1 (1 cpu)",
            "16 (16 cpus)"
        ])

    mem: ShapeArg[int] = ShapeArg[int](
        match = lambda arg: bool(re.match(r'^[0-9]+[MG]B?$', arg)),
        format = formatters.mem,
        value = "4G",
        name = "Memory",
        syntax="d(M|G)[B]",
        examples=[
            "3000MB (3 GB)",
            "16G (16GB)"
        ],
        help="Amount of memory to request"
    )

    x11: FlagArg = FlagArg(
        match=["x11", "--x11"],
        help="Request X-forwarding for GUI applications. Must also request X-forwarding on your SSH connection."
    )

    directory: ShapeArg[Path] = ShapeArg[Path](
        match = lambda arg: Path(arg).exists() and Path(arg).is_dir(),
        format = Path,
        value = ".",
        name = "Directory",
        syntax="/absolute/path | ./relative/path",
        help="Change working directory before submitting the command. All other "
             "relative paths will be evaluated relative to the new directory."
    ) 

    test: FlagArg = FlagArg(
        match=["-t", "--test"],
        help="Test mode. Will echo SLURM settings and the command without submitting anything."
    )

    job_template: KeywordArg[str] = KeywordArg[str](
        match=['-j', '--job-template'],
        validate=validators.job_template,
        num=1,
        values_name="template",
        help="Set CPU, memory, and time from a template. "
             "Run with -J to see list of templates. Any "
             "template item can be overridden by supplying "
             "the appropriate arg."
    )

    list_job_templates: FlagArg= FlagArg(
        match=["-J", "--list", "-l"],
        help="List all available templates."
    )
    
    account: KeywordArg[str] = KeywordArg[str](
        match=['-a', '--account'],
        num=1,
        values=['ctb-akhanf'],
        help="Compute account to submit the job under.",
        values_name="name"
    )

    help: FlagArg = FlagArg(
        match=['-h', '--help'],
        help="Print this help message"
    )
    
    tail: TailArg = TailArg("Command")






import re
import attr

from kslurm.args import ShapeArg, TailArg

VERSION_REGEX = (
    r"v?(\d+)(?:\.(\d+))?(?:\.(\d+))?(?:\.(\d+))?"
    "("
    "[._-]?"
    r"(?:(stable|beta|b|rc|RC|alpha|a|patch|pl|p)((?:[.-]?\d+)*)?)?"
    "([.-]?dev)?"
    ")?"
    r"(?:\+[^\s]+)?"
)

@attr.s(auto_attribs=True)
class UpdateModel:
    version: ShapeArg[str] = ShapeArg[str](
        match = lambda arg: bool(re.match(VERSION_REGEX, arg)),
        value = "",
        name="Version",
        syntax="d[.d][.d][.d][_tag]",
        examples=[
            "0.1.0"
        ],
        help="Optional: version to install"
    )

    tail: TailArg = TailArg()


from rich.theme import Theme

default = Theme({
    "heading": "green bold",
    "hot": "deep_pink3",
    "grey": "grey54",

    "syntax_col": "cyan bold",
    "examples_col": "grey54",
    "default_col": "deep_pink3",
    "values_col": "italic"
})


from typing import cast
import json
import re
import os
import site
import sys
import functools as ft
from pathlib import Path
from urllib.request import Request, urlopen
from contextlib import closing
from kslurm.models import VERSION_REGEX

def data_dir(home_dir_var: str) -> Path:
    dir = (Path(sys.executable) / "../../..").resolve()
    # Dir should have VERSION file
    if (dir / "VERSION").exists():
        return dir

    # If not, check if they have their HOME_DIR set 
    if os.getenv(home_dir_var):
        return Path(os.getenv(home_dir_var)).expanduser() # type: ignore

    # If still nothing, we'll just install at the usual place
    path = os.getenv("XDG_DATA_HOME", Path.home() / ".local/share")
    path = Path(path) / "kutils"

    return path

def bin_dir(home_dir_var: str) -> Path:
    if os.getenv(home_dir_var):
        return Path(os.getenv(home_dir_var), "bin").expanduser() # type: ignore

    user_base = site.getuserbase()

    bin_dir = os.path.join(user_base, "bin")

    return Path(bin_dir)

def get(url: str):
    request = Request(url, headers={"User-Agent": "Python kslurm"})

    with closing(urlopen(request)) as r:
        return r.read()

def get_version(requested_version: str, preview: bool, force: bool, data_dir: Path, metadata_url: str):
    version_regex = re.compile(VERSION_REGEX)
    current_version = None
    if data_dir.joinpath("VERSION").exists():
        current_version = data_dir.joinpath("VERSION").read_text().strip()

    metadata = json.loads(get(metadata_url).decode())

    def _compare_versions(x:str, y:str):
        mx = version_regex.match(x)
        my = version_regex.match(y)

        if mx and my:
            vx = tuple(int(p) for p in mx.groups()[:3]) + (mx.group(5),)
            vy = tuple(int(p) for p in my.groups()[:3]) + (my.group(5),)

            if vx < vy:
                return -1
            elif vx > vy:
                return 1

            return 0
        else:
            raise Exception("Could not match version information")

    print("")
    releases = sorted(
        metadata["releases"].keys(), key=ft.cmp_to_key(_compare_versions)
    )

    if requested_version and requested_version not in releases:
        print(
            f"Version {requested_version} does not exist."
        )

        return None

    version = requested_version
    if not version:
        for release in reversed(releases):
            m = version_regex.match(release)
            if m and m.group(5) and not preview:
                continue

            version = release

            break
    assert isinstance(version, str)
    if current_version == version and not force:
        print(
            f"The latest version ({version}) is already installed."
        )

        return None

    return cast(str, version)


from typing import Generic, List, TypeVar

import os

import kslurm.args as arglib
from kslurm.exceptions import TemplateError, ValidationError
from kslurm.slurm import job_templates as templates, helpers
from kslurm.args import ShapeArg
from kslurm.models import SlurmModel

T = TypeVar("T", bound=SlurmModel)


class SlurmCommand(Generic[T]):
    def __init__(self, args: List[str], model: T):
        self._name = ""
        self._output = ""

        try:
            parsed = arglib.parse_args(args, model)
        except TemplateError as err:
            print(err.msg)
            print("Choose from the following list of templates:\n")
            templates.list_templates()
            exit()

        if parsed.list_job_templates.value and not parsed.help.value:
            templates.list_templates()
            exit()

        # set_template returns templated values only if a template is passed
        # if we pass a blank string, the models are returned unchanged
        template = parsed.job_template.values[0] if parsed.job_template.value else ""
        template_vals = templates.set_template(
            template, mem=model.mem, cpu=model.cpu, time=model.time
        )

        # Start by setting these three with model/template
        self.time = template_vals.time
        self.cpu = template_vals.cpu
        self.mem = template_vals.mem

        # Then update if values were specifically supplied on the command line
        if parsed.time.updated:
            self.time = parsed.time
        if parsed.cpu.updated:
            self.cpu = parsed.cpu
        if parsed.mem.updated:
            self.mem = parsed.mem

        self.gpu = bool(parsed.gpu.value)
        self.x11 = bool(parsed.x11.value)
        self.account = parsed.account
        self.cwd = parsed.directory
        self.test = bool(parsed.test.value)
        self.job_template = parsed.job_template
        self._command = parsed.tail.values
        self.help = bool(parsed.help.value)

        os.chdir(self.cwd.value) # type: ignore

        self.script = [self.command]

        self.args = parsed

    ###
    # Job Paramaters
    ###
    @property
    def time(self):
        return helpers.slurm_time_format(self._time.value) # type: ignore

    @time.setter
    def time(self, time: ShapeArg[int]):
        self._time = time

    @property
    def name(self):
        if not self._name:
            return self._command[0]
        else:
            return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, output: str):
        self._output = f'--output="{output}"'
    
    ###
    # Command line strings
    ###
    @property
    def slurm_args(self):
        s = f"--account={self.account} --time={self.time} --cpus-per-task={self.cpu} --mem={self.mem}"
        if self.gpu:
            s += " --gres=gpu:1"
        if self.x11:
            s += " --x11"
        return s

    @property
    def command(self):
        return " ".join(self._command)

    @command.setter
    def command(self, commands: List[str]):
        self._command = commands

    @property
    def script(self):
        return self._script

    @script.setter
    def script(self, command: List[str]):
        self._script = "\n".join(["#!/bin/bash"] + command)

    ###
    # Job submission commands
    ###
    @property
    def run(self):
        if self.command:
            return f"echo '{self.command}' | srun {self.slurm_args} bash"
        else:
            return f"salloc {self.slurm_args}"

    @property
    def batch(self):
        if self.test:
            s = "cat"
        else:
            s = f"sbatch {self.slurm_args} --job-name={self.name} --parsable {self.output}"
        if self.command:
            return f"echo '{self.script}' | {s}"
        else:
            raise ValidationError("No command given")

    

    

    


import attr

from kslurm.args import ChoiceArg, TailArg

@attr.s(auto_attribs=True)
class KslurmModel:
    command: ChoiceArg[str] = ChoiceArg[str](
        match=[
            "kbatch",
            "krun",
            "kjupyter",
            "update"
        ],
        name="Command",
        help="Run any of the commands with -h to get more information"
    )

    tail: TailArg = TailArg("Args")


from typing import List
from kslurm.models import UpdateModel
from kslurm.args import parse_args
from kslurm.installer.utils import data_dir, bin_dir, get_version
from kslurm.installer.install import install

METADATA_URL = "https://pypi.org/pypi/kslurm/json"
NAME = "kslurm"
HOME_DIR = "KSLURM_HOME"
ENTRYPOINTS = [
    "kbatch",
    "krun",
    "kjupyter",
    "kslurm"
]

def update(args: List[str]):
    parsed = parse_args(args, UpdateModel())
    data = data_dir(HOME_DIR)
    bin = bin_dir(HOME_DIR)
    print("Checking for Updates")
    version = get_version(parsed.version.value, False, False, data, METADATA_URL)
    if version is None:
        return 1
    specification = f"kslurm=={version}"

    return install(version, specification, data, bin, True, ENTRYPOINTS)

    
if __name__ == "__main__":
    update([])


from typing import Any, Iterable, List, Optional, TypeVar, Union, cast

import textwrap
from pathlib import Path

from rich.table import Table
from rich.console import Group
from rich.padding import Padding
from rich.text import Text

from .arg_types import Arg, ChoiceArg, FlagArg, KeywordArg, PositionalArg, ShapeArg, TailArg
from .helpers import group_by_type, get_arg_list
from kslurm.style.console import console

def _syntax_format(syntax: str):
    lines = [f"[cyan bold]{line.strip()}[/]" for line in syntax.split(" | ")]
    return "\n".join(lines)

def _header(header: str): 
    return f"[bold green]{header.upper()}[/bold green]"

def _section(header: str, body: Union[Table, str]):
    return Group(
        Padding.indent(_header(header), 4),
        Padding.indent(body, 8),
        "\n"
    ) if body else ""

def print_help(script: str, models: object, script_help: str = "") -> None:
    arg_list = get_arg_list(models)
    grouped = group_by_type(arg_list)
    positionals =  cast(
        List[PositionalArg[Any]], 
        grouped.get(PositionalArg, [])
    )
    shapes = cast(
        List[ShapeArg[Any]],
        grouped.get(ShapeArg, [])
    )
    keywords = cast(
        List[KeywordArg[Any]],
        grouped.get(KeywordArg, [])
    )
    flags = cast(
        List[FlagArg],
        grouped.get(FlagArg, [])
    )
    choices = cast(
        List[ChoiceArg[Any]],
        grouped.get(ChoiceArg, [])
    )
    tail = cast(
        List[TailArg],
        grouped.get(TailArg, [])
    )
    positional_names = [arg.name for arg in positionals]
    script_name = Path(script).name
    
    command_line_example = f"[b]USAGE:[/] {script_name} [hot]<keywords and flags>[/] {' '.join(positional_names)}"
    if tail:
        command_line_example += f" [cyan]<{tail[0].name.lower()}>[/]"
    command_line_example += "\n\n"

    shape_section = _section('Shape args', _shape_table(shapes))
    keyword_section = _section('keyword args', _keyword_table(keywords))
    positional_section = _section('positional args', _positional_table(positionals))
    flag_section = _section('flag args', _flag_table(flags))
    choice_section = _section('choice args', _choice_table(choices))
    sections = list(filter(None, [
        "\n",
        command_line_example,
        script_help,
        choice_section,
        positional_section,
        shape_section,
        keyword_section,
        flag_section
    ]))
    console.print(*sections)
    
    

T = TypeVar("T", bound=Arg[Any])

def _get_helps(args: List[T]):
    return [Text(textwrap.fill(arg.help, 70)) for arg in args]


def _get_defaults(args: List[Any]):
    return [Text(str(arg), style="default_col") for arg in args]


def _help_table(
        rows: Iterable[Iterable[Union[Text,str]]], 
        header: Optional[List[str]]=None,
        right_aligned_cols: int = 1):
    t = Table.grid(padding=(1, 2), )
    for _ in range(right_aligned_cols):
        t.add_column(justify="right")
    if header:
        t.add_row(*header, style="bold")
    for r in rows:
        t.add_row(*list(r))
    return t

def _shape_table(args: List[ShapeArg[Any]]):
    if args:
        names =    [ Text(arg.name, style="bold")                for arg in args]
        syntaxes = [ _syntax_format(arg.syntax)                  for arg in args]
        examples = [ Text("\n".join(arg.examples), style="grey") for arg in args]
        eg =       [ "➔" if example else ""                     for example in examples ]
        defaults =   _get_defaults([arg.raw_value                for arg in args])
        helps =      _get_helps(args)

        header = ["", "Syntax", "", "Examples", "Default", ""]
        body = zip(names, syntaxes, eg, examples, defaults, helps)
            
        return _help_table(body, header, right_aligned_cols=2)
    else:
        return ""

def _positional_table(args: List[PositionalArg[Any]]):
    if args:
        names = [Text(arg.name, style="bold") for arg in args]
        defaults = _get_defaults([arg.raw_value for arg in args])
        helps = _get_helps(args)
        body = zip(names, defaults, helps)
        header = ["", "Default", ""]
        return _help_table(body, header)
    else:
        return ""

def _keyword_table(args: List[KeywordArg[Any]]):
    if args:
        names = [Text(', '.join(arg.match_list), style="bold") for arg in args]
        value_names = [f"<[i]{arg.values_name}[/]>" for arg in args]
        defaults = _get_defaults([arg.values for arg in args])
        helps = _get_helps(args)
        
        body = zip(names, value_names, defaults, helps)
        header = ["", "", "Default", ""]
        return _help_table(body, header)
    else:
        return ""

def _flag_table(args: List[FlagArg]):
    if args:
        names = [Text(', '.join(arg.match_list), style='bold') for arg in args]
        helps = _get_helps(args)
        
        body = zip(names, helps)
        return _help_table(body)
    else:
        return ""

def _choice_table(args: List[ChoiceArg[Any]]):
    if args:
        names = [Text(arg.name, style="bold") for arg in args]
        choices = [Text(', '.join(arg.match_list), style='bold') for arg in args]
        helps = _get_helps(args)
        defaults = _get_defaults([arg.raw_value for arg in args])

        body = zip(names, choices, defaults, helps)
        return _help_table(body)
    else:
        return ""


from __future__ import annotations
from typing import Optional

class CommandLineError(Exception):
    def __init__(self, msg: str, src_err: Optional[ValidationError] = None):
        super().__init__(msg)
        self.msg = msg
        self.src_err = src_err

class ValidationError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)
        self.msg = msg

class TemplateError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)
        self.msg = msg


import re

def time(time: str):
    if ':' in time:
        if '-' in time:
            days, hhmm = time.split('-')
            hours, min = hhmm.split(':')
        else:
            days = 0
            hours, min = time.split(':')
    else:
        try:
            min = int(time)
            hours, days = 0, 0
        except:
            raise TypeError(f"Invalid format for time: \"{time}\"\n"
                            f"Must be as [xx-]xx:xx or x where x is a number")
    return int(min) + int(hours)*60 + int(days)*60*24

def mem(mem: str):
    match = re.match(r'^[0-9]+', mem)
    if match:
        num = int(match.group())
    else:
        raise Exception("Memory is not formatted correctly")
    if 'G' in mem:
        return num * 1000
    else:
        return num



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kslurm',
 'kslurm.args',
 'kslurm.installer',
 'kslurm.models',
 'kslurm.slurm',
 'kslurm.style',
 'kslurm.submission']

package_data = \
{'': ['*'], 'kslurm': ['data/*']}

install_requires = \
['attrs>=21.2.0,<22.0.0',
 'colorama>=0.4.4,<0.5.0',
 'rich>=10.9.0,<11.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'typing-extensions>=3.10,<4.0']

entry_points = \
{'console_scripts': ['kbatch = kslurm.submission.kbatch:main',
                     'kjupyter = kslurm.submission.kjupyter:main',
                     'krun = kslurm.submission.krun:main',
                     'kslurm = kslurm.kslurm:main']}

setup_kwargs = {
    'name': 'kslurm',
    'version': '0.2.3',
    'description': 'Helper scripts and wrappers for running commands on SLURM compute clusters.',
    'long_description': "Utility functions to make working with SLURM easier. \n\n# Installation\nCluster utils is meant to be run in a SLURM environment, and thus will only install on linux. Open a shell and run the following command:\n\n```\ncurl -sSL https://raw.githubusercontent.com/pvandyken/kslurm/master/install_kslurm.py | python -\n```\n\nIf you wish to uninstall, run the same command with --uninstall added to the end.\n\nThe package can be updated by running `kslurm update`. \n\n# Features\nCurrently offers three commands:\n* kbatch: for batch submission jobs (no immediate output)\n* krun: for interactive submission\n* kjupyter: for Jupyter sessions\n\nAll three use a regex-based argument parsing, meaning that instead of writing a SLURM file or supplying confusing `--arguments`, you can request resources with an intuitive syntax:\n\n```\nkrun 4 3:00 15G gpu \n```\nThis command will request interactive session with __4__ cores, for __3hr__, using __15GB__ of memory, and a __gpu__.\n\nAnything not specfically requested will fall back to a default. For instance, by default the commands will request 3hr jobs using 1 core with 4GB of memory. You can also run a predefined job template using -j _template_. Run either command with -J to get a list of all templates. Any template values can be overriden simply by providing the appropriate argument.\n\nThe full list of possible requests, their syntaxes, and their defaults can be found at the bottom of the README.\n\n## krun\n\nkrun is used for interactive sessions on the cluster. If you run krun all by itself, it will fire up an interactive session on the cluster:\n\n```\nkrun\n```\nYou'll notice the server name in your terminal prompt will be changed to the cluster assigned to you. To end the session, simply use the `exit` command.\n\nYou can also submit a specific program to run:\n\n```\nkrun 1:00 1G python my_program.py\n```\nThis will request a 1hr session with one core and 1 GB of memory. The output of the job will be displayed on the console. Note that your terminal will be tied to the job, so if you quit, or get disconnected, your job will end. (tmux can be used to help mitigate this, see this [tutorial from Yale](https://docs.ycrc.yale.edu/clusters-at-yale/guides/tmux/) for an excellent overview).\n\nNote that you should never request more than the recommended amount of time for interactive jobs as specified by your cluster administrator. For ComputeCanada servers, you should never request more than 3 hr. If you do, you'll be placed in the general pool for resource assignment, and the job could take hours to start. Jobs of 3hr or less typically start in less than a minute.\n\n## kbatch\n\nJobs that don't require monitoring of the output or immediate submission, or will run for more than three hours, should be submitted using `kbatch`. This command schedules the job, then returns control of the terminal. Output from the job will be placed in a file in your current working directory entitled `slurm-[jobid].out`. \n\nImproving on `sbatch`, `kbatch` does not require a script file. You can directly submit a command:\n\n```\nkbatch 2-00:00 snakemake --profile slurm\n```\nThis will schedule a 2 day job running snakemake.\n\nOf course, complicated jobs can still be submitted using a script. Note that kbatch explictely specifies the resources it knows about in the command line. Command line args override `#SBATCH --directives` in the submit script, so at this time, you cannot use such directives to request resources unless they are not currently supported by kslurm. This may change in a future release.\n\n## kjupyter\n\nThis command requests an interactive job running a jupyter server. As with krun, you should not request a job more than the recommended maximum time for your cluster (3hr for ComputeCanada). If you need more time than that, just request a new job when the old one expires.\n\nYou should not provide any extra command to kjupyter. Just supply whatever resources you wish to request.\n```\nkjupyter 32G 2\n```\nThis will start a jupyter session with 32 GB of memory and 2 cores.\n\nNote that the command will fail if there is no `jupyter-lab` executable on the `PATH`. Use `pip install jupyterlab` if it's not installed. Typically, you should do this within a Python environment using a tool of your choice (e.g. virtualenv).\n\n# Unsupported SLURM args\n\nCurrently, the only way to supply arguments to SLURM beyond the items listed below is to list it as an `#SBATCH --directive` in a submission script. This only works with `kbatch`, not `krun` or `kjupyter`. A future release may support a method to supply these arguments directly on the command line. If you frequently use an option not listed below, make an issue and we can discuss adding support!\n\n# Slurm Syntax\n\nThe full syntax is outlined below. You can always run a command with `-h` to get help.\n\n| Resource  |           Syntax           |                                                                   Default |                                                    Description |\n| :-------- | :------------------------: | ------------------------------------------------------------------------: | -------------------------------------------------------------: |\n| Time      | [d-]dd:dd -> [days-]hh:mm  |                                                                       3hr |                                   The amount of time requested |\n| CPUS      |     d -> just a number     |                                                                         1 |                                   The number of CPUs requested |\n| Memory    |            d(G/M)[B] -> e.g. 4G, 500MB |                                                            4GB | The amount of memory requested |\n| Account   | --account <_account name_> | Currently hard coded to ctb-akhanf. Eventually will support configuration |                      The account under which to submit the job |\n| GPU       |            gpu             |                                                                     False |                         Provide flag to request 1 GPU instance |\n| Directory |  <_any valid directory_>   |                                                                        ./ | Change the current working directory before submitting the job |\n| x11       |           --x11            |                                                                     False |                   Requests x11 forwarding for GUI applications |\n",
    'author': 'Peter Van Dyken',
    'author_email': 'pvandyk2@uwo.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pvandyken/kslurm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
